/* Copyright (C) 2003-2015 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Martin Schwidefsky <schwidefsky@de.ibm.com>, 2003.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */

#include <endian.h>
#include <errno.h>
#include <sysdep.h>
#include <lowlevellock.h>
#include <pthread.h>
#include <pthreadP.h>
#include <stap-probe.h>
#include <atomic.h>

#include <shlib-compat.h>
#include <kernel-features.h>


/* See __pthread_cond_wait for a high-level description of the algorithm.  */
int
__pthread_cond_broadcast (pthread_cond_t *cond)
{
  unsigned int gen, wseq, ssent;

  /* See comment in __pthread_cond_signal.  */
  int pshared = (atomic_load_relaxed (&cond->__data.__mutex) == (void *) ~0l)
		? LLL_SHARED : LLL_PRIVATE;

  LIBC_PROBE (cond_broadcast, 1, cond);

  /* We use the same approach for broadcasts as for normal signals but wake
     all waiters (i.e., we try to set SIGNALS_SENT to WSEQ).  However, to
     prevent an excessive number of spurious wake-ups, we need to check
     whether we read values for SIGNALS_SENT and WSEQ that are from one
     generation; otherwise, we could read a not-yet-reset WSEQ and a reset
     SIGNAL_SENT, resulting in a very large number of spurious wake-ups that
     we make available.  Checking the generation won't prevent an ABA problem
     for the CAS in the loop below when the generation changes between our
     generation check and the CAS; however, in this case, we just add a still
     reasonable number of spurious wake-ups (i.e., equal to the number of
     waiters than were actually blocked on the condvar at some point in the
     past).  Therefore, we first load the current generation.  We need
     acquire MO here to make sure that we next read values for WSEQ and
     SIGNALS_SENT from this or a later generation (see the matching release
     MOs in __pthread_cond_wait).  */
  gen = atomic_load_acquire (&cond->__data.__generation);
  wseq = atomic_load_relaxed (&cond->__data.__wseq);
  ssent = atomic_load_relaxed (&cond->__data.__signals_sent);
  do
    {

      /* If the generation changed concurrently, then we could have been
         positioned in the earlier generation; thus, all waiters we must wake
         have been or will be woken during the quiescence period.  The other
         conditions are the same as in __pthread_cond_signal.
         We add an acquire-MO fence to ensure that loading the stores to WSEQ
         and SIGNALS_SENT that we read from above happened before we read
         GENERATION a second time, which allows us to detect if we read
         partially reset state or state from a new generation (see
         __pthread_cond_wait and the matching release MO stores there).  */
      atomic_thread_fence_acquire ();
      if (gen != atomic_load_relaxed (&cond->__data.__generation)
	  || ssent >= wseq || wseq >= __PTHREAD_COND_WSEQ_THRESHOLD)
	return 0;
    }
  while (!atomic_compare_exchange_weak_relaxed (&cond->__data.__signals_sent,
						&ssent, wseq));

  /* XXX Could we skip the futex_wake if not necessary (eg, if there are just
     spinning waiters)?  This would need additional communication but could it
     be more efficient than the kernel-side communication?  Should we spin for
     a while to see if our signal was consumed in the meantime?  */
  /* TODO Use futex_requeue on the mutex?  Could increase broadcast scalability
     if there are many waiters, but this depends on the scalability of the
     mutex.  It also prevents use of lock elision, and requires all waiters
     to put the mutex in contended state when they re-acquire it, independent
     of whether they were woken by a broadcast or not.  Notes that we can only
     requeue if the mutex is set already (ie, we had waiters already).  */
  /* XXX Really INT_MAX? Would WSEQ-SIGNALS_SENT be possible?  Useful?  */
  lll_futex_wake (&cond->__data.__signals_sent, INT_MAX, pshared);

  return 0;
}

versioned_symbol (libpthread, __pthread_cond_broadcast, pthread_cond_broadcast,
		  GLIBC_2_3_2);
