/* Copyright (C) 2002-2015 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Ulrich Drepper <drepper@redhat.com>, 2002.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */

#include <errno.h>
#include <shlib-compat.h>
#include "pthreadP.h"
#include <stap-probe.h>
#include <atomic.h>


/* See __pthread_cond_wait for a high-level description of the algorithm.

   A correct program must make sure that no waiters are blocked on the condvar
   when it is destroyed, and that there are no concurrent signals or
   broadcasts.  To wake waiters reliably, the program must signal or
   broadcast while holding the mutex or after having held the mutex.  It must
   also ensure that no signal or broadcast are still pending to unblock
   waiters; IOW, because waiters can wake up spuriously, the program must
   effectively ensure that destruction happens after the execution of those
   signal or broadcast calls.
   Thus, we can assume that any waiters that are still accessing the condvar
   will either (1) have been woken but not yet confirmed that they woke up or
   (2) wait for quiescence to finish (i.e., the only steps they will perform
   are waiting on GENERATION and then decrementing QUIESCENCE_WAITERS; all
   other steps related to quiescence are performed by waiters before they
   release the mutex).
   Thus, if we are not yet in an ongoing quiescence state, we just make
   the last concurrently confirming waiter believe we are so that it notifies
   us; then we wait for QUIESCENCE_WAITERS to finish waiting for the end of
   the quiescence state.  */
int
__pthread_cond_destroy (pthread_cond_t *cond)
{
  unsigned int wseq, val;

  /* See comment in __pthread_cond_signal.  */
  int pshared = (atomic_load_relaxed (&cond->__data.__mutex) == (void *) ~0l)
		? LLL_SHARED : LLL_PRIVATE;

  LIBC_PROBE (cond_destroy, 1, cond);

  /* If we are already in the quiescence state, then signals and broadcasts
     will not modify SIGNALS_SENT anymore because all waiters will wake up
     anyway (and we don't have to synchronize between signal/broadcast and the
     reset of SIGNALS_SENT when quiescence is finished).  Thus, never do the
     following check in this case; it cannot be reliably anyway, and is also
     just recommended by POSIX.  */
  wseq = atomic_load_relaxed (&cond->__data.__wseq);
  if (wseq != __PTHREAD_COND_WSEQ_THRESHOLD
      && wseq > atomic_load_relaxed (&cond->__data.__signals_sent))
    return EBUSY;

  /* Waiters can either be (1) pending to confirm that they have been woken
     or (2) spinning/blocking on GENERATION to become odd.  Thus, we first
     need to make sure that any waiter woken by the program has finished the
     condvar-internal synchronization (i.e., it has confirmed the wake-up).
     We use the quiescence mechanism to get notified when all of them are
     finished by adding the right amount of artificial confirmed waiters.
     XXX Or is just relaxed MO sufficient because happens-before is
     established through the total modification order on CONFIRMED?  */
  if (atomic_fetch_add_acq_rel (&cond->__data.__confirmed,
				__PTHREAD_COND_WSEQ_THRESHOLD - wseq)
      < wseq)
    {
      /* There are waiters that haven't yet confirmed.  If we have an even
	 number for generation, wait until it is changed by the last waiter
	 to confirm.  (The last waiter will increase to WSEQ_THRESHOLD, so
	 it will increase GENERATION to an odd value.)  We need acquire MO
	 to make any waiters' accesses to the condvar happen before we
	 destroy it.*/
      while (1)
	{
	  val = atomic_load_acquire (&cond->__data.__generation);
	  if ((val & 1) != 1)
	    lll_futex_wait (&cond->__data.__generation, val, pshared);
	  else
	    break;
	}
    }

  /* If we are in a quiescence period, we also need to wait for those waiters
     that are waiting for quiescence to finish.  Note that we cannot have
     pushed waiters into this state by artificially introducing quiescence
     above, so we also do not wake any such waiters.  As above, we need
     acquire MO.  */
  while (1)
    {
      val = atomic_load_acquire (&cond->__data.__quiescence_waiters);
      if (val > 0)
	lll_futex_wait (&cond->__data.__quiescence_waiters, val, pshared);
      else
	break;
    }

  /* The memory the condvar occupies can now be reused.  */
  return 0;
}
versioned_symbol (libpthread, __pthread_cond_destroy,
		  pthread_cond_destroy, GLIBC_2_3_2);
