/* Copyright (C) 2003-2015 Free Software Foundation, Inc.
   This file is part of the GNU C Library.
   Contributed by Martin Schwidefsky <schwidefsky@de.ibm.com>, 2003.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, see
   <http://www.gnu.org/licenses/>.  */

#include <endian.h>
#include <errno.h>
#include <sysdep.h>
#include <lowlevellock.h>
#include <pthread.h>
#include <pthreadP.h>
#include <sys/time.h>
#include <atomic.h>

#include <shlib-compat.h>
#include <stap-probe.h>
#include <kernel-features.h>

#ifndef HAVE_CLOCK_GETTIME_VSYSCALL
# undef INTERNAL_VSYSCALL
# define INTERNAL_VSYSCALL INTERNAL_SYSCALL
# undef INLINE_VSYSCALL
# define INLINE_VSYSCALL INLINE_SYSCALL
#else
# include <libc-vdso.h>
#endif

struct _condvar_cleanup_buffer
{
  int oldtype;
  pthread_cond_t *cond;
  pthread_mutex_t *mutex;
};

static __always_inline void
__condvar_confirm_wakeup (pthread_cond_t *cond, int pshared)
{
  /* Confirm that we have been woken.  If the number of confirmations reaches
     WSEQ_THRESHOLD, we must be in a quiescence period (i.e., WSEQ must be
     equal to WSEQ_THRESHOLD).
     We use acquire-release MO to ensure that accesses to this generation's
     condvar state happen before any reset of the condvar.
     XXX Or is just relaxed MO sufficient because happens-before is
     established through the total modification order on CONFIRMED?  */
  if (atomic_fetch_add_acq_rel (&cond->__data.__confirmed, 1)
      == __PTHREAD_COND_WSEQ_THRESHOLD - 1)
    {
      /* Need release MO to make our accesses to the condvar happen before
	 the reset that some other thread will execute.  */
      atomic_fetch_add_release (&cond->__data.__generation, 1);
      lll_futex_wake (&cond->__data.__generation, INT_MAX, pshared);
    }

}

/* Cancel waiting after having registered as a waiter already.
   We must not consume another waiter's signal, so we must add an artificial
   signal.  If we are the first blocked waiter (i.e., SEQ == SIGNALS_SENT,
   SEQ being our position in WSEQ), then an artificial signal is obviously
   fine.  If we are blocked (i.e., SEQ > SIGNALS_SENT), then a fake signal
   might lead to spurious wake-ups of waiters with a smaller position in WSEQ;
   however, not adding the artificial signal could prevent wake-up of waiters
   with a larger position in WSEQ because we weren't actually waiting yet
   effectively consume a signal because we have reserved a slot in WSEQ.  If
   we are not blocked anymore (i.e., SEQ < SIGNALS_SENT), we still have to
   add the artificial signal if there are still unblocked threads (i.e.,
   SIGNALS_SENT < WSEQ).  */
static __always_inline void
__condvar_cancel_waiting (pthread_cond_t *cond, int pshared)
{
  unsigned int wseq, ssent;

  /* Add an artificial signal.  See __pthread_cond_signal.  */
  wseq = atomic_load_relaxed (&cond->__data.__wseq);
  ssent = atomic_load_relaxed (&cond->__data.__signals_sent);
  do
    {
      if (ssent >= wseq || wseq >= __PTHREAD_COND_WSEQ_THRESHOLD)
	break;
    }
  while (!atomic_compare_exchange_weak_relaxed (&cond->__data.__signals_sent,
						&ssent, ssent + 1));
}

/* Clean-up for cancellation of waiters waiting for normal signals.  We cancel
   our registration as a waiter, confirm we have woken up, and re-acquire the
   mutex.  */
static void
__condvar_cleanup_waiting (void *arg)
{
  struct _condvar_cleanup_buffer *cbuffer =
    (struct _condvar_cleanup_buffer *) arg;
  pthread_cond_t *cond = cbuffer->cond;
  /* See comment in __pthread_cond_signal.  */
  int pshared = (atomic_load_relaxed (&cond->__data.__mutex) == (void *) ~0l)
		? LLL_SHARED : LLL_PRIVATE;

  __condvar_cancel_waiting (cond, pshared);
  __condvar_confirm_wakeup (cond, pshared);

  /* Cancellation can happen after we have been woken by a signal's
     futex_wake (unlike when we cancel waiting due to a timeout on futex_wait,
     for example).  We do provide an artificial signal in
     __condvar_cancel_waiting, but we still can have consumed a futex_wake
     that should have woken another waiter.  We cannot reliably wake this
     waiter because there might be other, non-eligible waiters that started
     to block after we have been cancelled; therefore, we need to wake all
     blocked waiters to really undo our consumption of the futex_wake.  */
  /* XXX Once we have implemented a form of cancellation that is just enabled
     during futex_wait, we can try to optimize this.  */
  lll_futex_wake (&cond->__data.__signals_sent, INT_MAX, pshared);

  /* XXX If locking the mutex fails, should we just stop execution?  This
     might be better than silently ignoring the error.  */
  __pthread_mutex_cond_lock (cbuffer->mutex);
}

/* Clean-up for cancellation of waiters waiting on quiescence to finish.  */
static void
__condvar_cleanup_quiescence (void *arg)
{
  struct _condvar_cleanup_buffer *cbuffer =
    (struct _condvar_cleanup_buffer *) arg;
  pthread_cond_t *cond = cbuffer->cond;
  /* See comment in __pthread_cond_signal.  */
  int pshared = (atomic_load_relaxed (&cond->__data.__mutex) == (void *) ~0l)
		? LLL_SHARED : LLL_PRIVATE;

  /* See __pthread_cond_wait.  */
  if (atomic_fetch_add_release (&cond->__data.__quiescence_waiters, -1) == 1)
    lll_futex_wake (&cond->__data.__quiescence_waiters, INT_MAX,
	pshared);

  /* XXX If locking the mutex fails, should we just stop execution?  This
     might be better than silently ignoring the error.  */
  __pthread_mutex_cond_lock (cbuffer->mutex);
}

/* This condvar implementation guarantees that all calls to signal and
   broadcast and all of the three virtually atomic parts of each call to wait
   (i.e., (1) releasing the mutex and blocking, (2) unblocking, and (3) re-
   acquiring the mutex) happen in some total order that is consistent with the
   happens-before relations in the calling program.  However, this order does
   not necessarily result in additional happens-before relations being
   established (which aligns well with spurious wake-ups being allowed).

   All waiters acquire a certain position in a waiter sequence, WSEQ.  Signals
   and broadcasts acquire a position or a whole interval in the SIGNALS_SENT
   sequence.  Waiters are allowed to wake up if either SIGNALS_SENT is larger
   or equal to their position in WSEQ, or if they have been blocked on a
   certain futex and selected by the kernel to wake up after a signal or
   broadcast woke threads that were blocked on this futex.  This is also the
   primary source for spurious wake-ups: For waiters W1 and W2 with W2's
   position in WSEQ larger than W1's, if W2 blocks earlier than W1 using this
   futex, then a signal will wake both W1 and W2.  However, having the
   possibility of waking waiters spuriously simplifies the algorithm and
   allows for a lean implementation.

   Futexes only compare 32b values when deciding whether to block a thread,
   but we need to distinguish more than 1<<32 states for the condvar.  Unlike
   mutexes, which are just free/acquired/contended, the position of waiters
   and signals matters because of the requirement of them forming a total
   order.  Therefore, to avoid ABA issues and prevent potential lost wake-ups,
   we need to safely reset WSEQ and SIGNALS_SENT.  We do so by quiescing the
   condvar once WSEQ reaches a threshold (WSEQ_THRESHOLD): We wait for all
   waiters to confirm that they have woken up by incrementing the CONFIRMED
   counter, and then reset the condvar state.  Waiters arriving in this
   quiescence period (i.e., the time between WSEQ reaching WSEQ_THRESHOLD and
   the reset being complete) will wake up spuriously.
   To avoid ABA issues for broadcasts that could lead to excessive numbers of
   spurious wake-ups, we maintain a GENERATION counter that increases
   whenever we enter and exit a quiescence period; waiters also use this
   counter to communicate when the quiescence period can be finished by
   incrementing GENERATION to an odd value.
   When waiters wait for quiescence to finish, they will have pending accesses
   to the condvar even though they are not registered as waiters.  Therefore,
   we count this number of waiters in QUIESCENCE_WAITERS; destruction of the
   condvar will not take place until there are no such waiters anymore.

   WSEQ is only modified while holding MUTEX, but signals and broadcasts read
   it without holding the mutex (see both functions for an explanation why
   this is safe).  SIGNALS_SENT is only modified with CAS operations by
   signals and broadcast; the only exception is the reset of the condvar
   during quiescence (but this is fine due to how signals and broadcasts
   update SIGNALS_SENT).  CONFIRMED is accessed by just waiters with atomic
   operations, and reset during quiescence.  GENERATION is modified by waiters
   during quiescence handling, and used by broadcasts to check whether a
   snapshot of WSEQ and SIGNALS_SENT happened within a single generation.
   QUIESCENCE_WAITERS is only modified by waiters that wait for quiescence to
   finish.

   The common-case state is WSEQ < WSEQ_THRESHOLD and GENERATION being even.
   CONFIRMED is always smaller or equal to WSEQ except during reset.
   SIGNALS_SENT can be larger than WSEQ, but this happens just during reset
   or if a signal or broadcast tripped over a reset or the hardware reordered
   in a funny way, in which case we just get a few more spurious wake-ups
   (see __pthread_cond_broadcast for details on how we minimize that).
   If WSEQ equals WSEQ_THRESHOLD, then incoming waiters will wait for all
   waiters in the current generation to finish, or they will reset the condvar
   and start a new generation.  If GENERATION is odd, the condvar state is
   ready for being reset.

   Limitations:
   * This condvar isn't designed to allow for more than
     WSEQ_THRESHOLD * (1 << (sizeof(GENERATION) * 8 - 1)) calls to
     __pthread_cond_wait.  It probably only suffers from potential ABA issues
     afterwards, but this hasn't been checked nor tested.
   * More than (1 << (sizeof(QUIESCENCE_WAITERS) * 8) -1 concurrent waiters
     are not supported.
   * Beyond what is allowed as errors by POSIX or documented, we can also
     return the following errors:
     * EPERM if MUTEX is a recursive mutex and the caller doesn't own it.
     * EOWNERDEAD or ENOTRECOVERABLE when using robust mutexes.  Unlike
       for other errors, this can happen when we re-acquire the mutex; this
       isn't allowed by POSIX (which requires all errors to virtually happen
       before we release the mutex or change the condvar state), but there's
       nothing we can do really.
     * EAGAIN if MUTEX is a recursive mutex and trying to lock it exceeded
       the maximum number of recursive locks.  The caller cannot expect to own
       MUTEX.
     * When using PTHREAD_MUTEX_PP_* mutexes, we can also return all errors
       returned by __pthread_tpp_change_priority.  We will already have
       released the mutex in such cases, so the caller cannot expect to own
       MUTEX.

   Other notes:
   * Instead of the normal mutex unlock / lock functions, we use
     __pthread_mutex_unlock_usercnt(m, 0) / __pthread_mutex_cond_lock(m)
     because those will not change the mutex-internal users count, so that it
     can be detected when a condvar is still associated with a particular
     mutex because there is a waiter blocked on this condvar using this mutex.
*/
static __always_inline int
__pthread_cond_wait_common (pthread_cond_t *cond, pthread_mutex_t *mutex,
    const struct timespec *abstime)
{
  struct _pthread_cleanup_buffer buffer;
  struct _condvar_cleanup_buffer cbuffer;
  const int maxspin = 0;
  int err;
  int result = 0;
  unsigned int spin, seq, gen, ssent;

  /* We (can assume to) hold the mutex, so there are no concurrent
     modifications.  */
  int pshared = (atomic_load_relaxed (&cond->__data.__mutex) == (void *) ~0l)
		? LLL_SHARED : LLL_PRIVATE;

  LIBC_PROBE (cond_wait, 2, cond, mutex);

  /* Remember the mutex we are using here, unless it's a pshared condvar.
     Users must ensure that a condvar is associated with exactly one mutex,
     so we cannot store an incorrect address if the program is correct.  */
  if (pshared != LLL_SHARED)
    atomic_store_relaxed (&cond->__data.__mutex, mutex);

  /* Acquire a position (SEQ) in the waiter sequence (WSEQ) iff this will not
     cause overflow.  We use an an atomic operation because signals and
     broadcasts may read while not holding the mutex.  We do not need release
     MO here because we do not need to establish any happens-before relation
     with signalers (see __pthread_cond_signal).  */
  seq = atomic_load_relaxed (&cond->__data.__wseq);
  if (__glibc_likely (seq < __PTHREAD_COND_WSEQ_THRESHOLD))
    atomic_store_relaxed (&cond->__data.__wseq, seq + 1);

  /* If we reached WSEQ_THRESHOLD, we need to quiesce the condvar.  */
  if (seq >= __PTHREAD_COND_WSEQ_THRESHOLD - 1)
    {
      /* If we are the waiter that triggered quiescence, we need to still
         confirm that we have woken up (which can update GENERATION if we are
         the last one active).
         XXX We probably do not need to wake anyone because we still hold the
         mutex so no other waiter can observe that we started quiescence.  */
      if (seq == __PTHREAD_COND_WSEQ_THRESHOLD - 1)
	__condvar_confirm_wakeup (cond, pshared);
      /* Check whether all waiters in the current generation have confirmed
	 that they do not wait anymore (and thus don't use the condvar state
	 anymore), and either reset or wait for this to happen.  We do that
	 while holding the mutex so we will never see WSEQ==WSEQ_THRESHOLD and
	 an even value for GENERATION that is already a new generation.  We
	 need acquire MO on the load to ensure that we happen after the last
	 of the current generation's waiters confirmed that it isn't using the
	 condvar anymore (see below).
	 Note that in both cases, we must not actually wait for any signal to
	 arrive but wake up spuriously.  This allows signalers to not take
	 actively part in quiescence because they can assume that if they
	 hit a quiescence period, all waiters they might have to wake will
	 wake up on their own.  */
      gen = atomic_load_acquire (&cond->__data.__generation);
      if ((gen & 1) != 0)
	{
	  /* No waiter uses the condvar currently, so we can reset.
	     This barrier / release-MO fence is necessary to match the
	     acquire-MO fence in __pthread_cond_broadcast.  It makes sure that
	     if a broadcast sees one of the values stored during reset, it
	     will also observe an even value for GENERATION (i.e., broadcast
	     can check whether it read condvar state that was from different
	     generations or partially reset).  We store atomically because
	     the fence, according to the memory model, only has the desired
	     effect in combination with atomic operations.  */
	  atomic_thread_fence_release ();
	  atomic_store_relaxed (&cond->__data.__wseq, 0);
	  atomic_store_relaxed (&cond->__data.__signals_sent, 0);
	  atomic_store_relaxed (&cond->__data.__confirmed, 0);
	  /* Need release MO to make sure that if a broadcast loads the new
	     generation, it will also observe a fully reset condvar.  */
	  atomic_fetch_add_release (&cond->__data.__generation, 1);
	  /* TODO Discuss issues around PI support on quiescence.  */
	  lll_futex_wake (&cond->__data.__generation, INT_MAX, pshared);
	  /* We haven't released the mutex, so we can just return.  */
	  return 0;
	}
      else
	{
	  /* There are still some waiters that haven't yet confirmed to not be
	     using the condvar anymore.  Wake all of them if this hasn't
	     happened yet.  Relaxed MO is sufficient because we only need to
	     max out SIGNALS_SENT and we still hold the mutex, so a new
	     generation cannot have been started concurrently.  */
	  ssent = atomic_load_relaxed (&cond->__data.__signals_sent);
	  while (1)
	    {
	      if (ssent == __PTHREAD_COND_WSEQ_THRESHOLD)
		break;
	      if (atomic_compare_exchange_weak_relaxed (
		  &cond->__data.__signals_sent, &ssent,
		  __PTHREAD_COND_WSEQ_THRESHOLD))
		{
		  /* If we made any signals available, wake up all waiters
		     blocked on the futex.  */
		  lll_futex_wake (&cond->__data.__signals_sent, INT_MAX,
				  pshared);
		  break;
		}
	    }
	  /* Now wait until no waiter is using the condvar anymore, and wake
	     up spuriously.  Don't hold the mutex while we wait.  We also
	     need to tell __pthread_cond_destroy that we will have pending
	     accesses to the condvar state; we do so before we release the
	     mutex to make sure that this is visible to destruction.  */
	  atomic_fetch_add_relaxed (&cond->__data.__quiescence_waiters, 1);
	  err = __pthread_mutex_unlock_usercnt (mutex, 0);

	  if (__glibc_likely (err == 0))
	    {
	      /* Enable asynchronous cancellation before we block, as required
		 by the standard.  In the cancellation handler, we just do
		 the same steps as after a normal futex wake-up.  */
	      cbuffer.cond = cond;
	      cbuffer.mutex = mutex;
	      __pthread_cleanup_push (&buffer, __condvar_cleanup_quiescence,
		  &cbuffer);
	      cbuffer.oldtype = __pthread_enable_asynccancel ();

	      /* We don't really need to care whether the futex_wait fails
		 because a spurious wake-up is just fine.  */
	      /* TODO Spin on generation (with acquire MO)?  */
	      /* TODO Discuss issues around PI support on quiescence.  */
	      lll_futex_wait (&cond->__data.__generation, gen, pshared);

	      /* Stopped blocking; disable cancellation.  */
	      __pthread_disable_asynccancel (cbuffer.oldtype);
	      __pthread_cleanup_pop (&buffer, 0);
	    }
	  /* Notify __pthread_cond_destroy that we won't access the condvar
	     anymore.  Release MO to make our accesses happen before
	     destruction.  */
	  if (atomic_fetch_add_release (&cond->__data.__quiescence_waiters, -1)
	      == 1)
	    lll_futex_wake (&cond->__data.__quiescence_waiters, INT_MAX,
		pshared);

	  /* If unlocking the mutex returned an error, we haven't released it.
	     We have decremented QUIESCENCE_WAITERS already, so we can just
	     return here.  */
	  if (__glibc_unlikely (err != 0))
	    return err;

	  /* Re-acquire the mutex, and just wake up spuriously.  */
	  /* XXX Rather abort on errors that are disallowed by POSIX?  */
	  return __pthread_mutex_cond_lock (mutex);
	}
    }

  /* Now that we are registered as a waiter, we can release the mutex.
     Waiting on the condvar must be atomic with releasing the mutex, so if
     the mutex is used to establish a happens-before relation with any
     signaler, the waiter must be visible to the latter; thus, we release the
     mutex after registering as waiter.
     If releasing the mutex fails, we just cancel our registration as a
     waiter and confirm that we have woken up.  */
  err = __pthread_mutex_unlock_usercnt (mutex, 0);
  if (__glibc_unlikely (err != 0))
    {
      __condvar_cancel_waiting (cond, pshared);
      __condvar_confirm_wakeup (cond, pshared);
      return err;
    }

  /* We might block on a futex, so push the cancellation handler.  */
  cbuffer.cond = cond;
  cbuffer.mutex = mutex;
  __pthread_cleanup_push (&buffer, __condvar_cleanup_waiting, &cbuffer);

  /* Loop until we might have been woken, which is the case if either (1) more
     signals have been sent than what is our position in the waiter sequence
     or (2) the kernel woke us after we blocked in a futex_wait operation.  We
     have to consider the latter independently of the former because the
     kernel might wake in an order that is different from the waiter sequence
     we determined (and we don't know in which order the individual waiters'
     futex_wait calls were actually processed in the kernel).
     We do not need acquire MO for the load from SIGNALS_SENT because we do
     not need to establish a happens-before with the sender of the signal;
     because every wake-up could be spurious, the program has to check its
     condition associated with the condvar anyway and must use suitable
     synchronization to do so.  IOW, we ensure that the virtual ordering of
     waiters and signalers is consistent with happens-before, but we do not
     transfer this order back into happens-before.  Also see the comments
     in __pthread_cond_signal.  */
  ssent = atomic_load_relaxed (&cond->__data.__signals_sent);
  spin = maxspin;
  while (ssent <= seq)
    {
      if (spin > 0)
	spin--;
      else
	{
	  if (abstime == NULL)
	    {
	      /* Enable asynchronous cancellation before we block, as required
		 by the standard.  */
	      cbuffer.oldtype = __pthread_enable_asynccancel ();
	      /* Block using SIGNALS_SENT as futex.  If we get woken due to a
		 concurrent change to the number of signals sent (i.e.,
		 EAGAIN is returned), we fall back to spinning and
		 eventually will try to block again.  All other possible
		 errors returned from the futex_wait call are either
		 programming errors, or similar to EAGAIN (i.e., EINTR
		 on a spurious wake-up by the futex).  Otherwise, we have
		 been woken by a real signal, so the kernel picked us for the
		 wake-up, and we can stop waiting.  */
	      err = lll_futex_wait (&cond->__data.__signals_sent, ssent,
				    pshared);
	      /* Stopped blocking; disable cancellation.  */
	      __pthread_disable_asynccancel (cbuffer.oldtype);
	      if (err == 0)
		break;
	    }
	  else
	    {
	      /* Block, but with a timeout.  */
	      /* Work around the fact that the kernel rejects negative timeout
		 values despite them being valid.  */
	      if (__glibc_unlikely (abstime->tv_sec < 0))
	        goto timeout;

#if (!defined __ASSUME_FUTEX_CLOCK_REALTIME \
     || !defined lll_futex_timed_wait_bitset)
	      struct timespec rt;
	      {
# ifdef __NR_clock_gettime
		INTERNAL_SYSCALL_DECL (err);
		(void) INTERNAL_VSYSCALL (clock_gettime, err, 2,
		    cond->__data.__clockid, &rt);
		/* Convert the absolute timeout value to a relative
		   timeout.  */
		rt.tv_sec = abstime->tv_sec - rt.tv_sec;
		rt.tv_nsec = abstime->tv_nsec - rt.tv_nsec;
# else
		/* Get the current time.  So far, we support only one
		   clock.  */
		struct timeval tv;
		(void) __gettimeofday (&tv, NULL);
		/* Convert the absolute timeout value to a relative
		   timeout.  */
		rt.tv_sec = abstime->tv_sec - tv.tv_sec;
		rt.tv_nsec = abstime->tv_nsec - tv.tv_usec * 1000;
# endif
	      }
	      if (rt.tv_nsec < 0)
		{
		  rt.tv_nsec += 1000000000;
		  --rt.tv_sec;
		}
	      /* Did we already time out?  */
	      if (__glibc_unlikely (rt.tv_sec < 0))
		goto timeout;

	      /* Enable asynchronous cancellation before we block, as required
		 by the standard.  */
	      cbuffer.oldtype = __pthread_enable_asynccancel ();
	      err = lll_futex_timed_wait (&cond->__data.__signals_sent, ssent,
		  &rt, pshared);

#else
	      unsigned int clockbit = (cond->__data.__clockid == 1
		  ? 0 : FUTEX_CLOCK_REALTIME);
	      /* Enable asynchronous cancellation before we block, as required
		 by the standard.  */
	      cbuffer.oldtype = __pthread_enable_asynccancel ();
	      err = lll_futex_timed_wait_bitset (&cond->__data.__signals_sent,
		  ssent, abstime, clockbit, pshared);
#endif
	      /* Stopped blocking; disable cancellation.  */
	      __pthread_disable_asynccancel (cbuffer.oldtype);

	      if (err == 0)
		break;
	      else if (__glibc_unlikely (err == -ETIMEDOUT))
		{
		  timeout:
		  /* When we timed out, we effectively cancel waiting.  */
		  __condvar_cancel_waiting (cond, pshared);
		  result = ETIMEDOUT;
		  break;
		}
	    }

	  spin = maxspin;
	}

      /* (Spin-)Wait until enough signals have been sent.  */
      ssent = atomic_load_relaxed (&cond->__data.__signals_sent);
    }

  /* We won't block on a futex anymore.  */
  __pthread_cleanup_pop (&buffer, 0);

  /* Confirm that we have been woken.  We do that before acquiring the mutex
     to reduce the latency of dealing with quiescence, and to allow that
     pthread_cond_destroy can be executed while having acquired the mutex.
     Neither signalers nor waiters will wait for quiescence to complete
     while they hold the mutex.  */
  __condvar_confirm_wakeup (cond, pshared);

  /* Woken up; now re-acquire the mutex.  If this doesn't fail, return RESULT,
     which is set to ETIMEDOUT if a timeout occured, or zero otherwise.  */
  err = __pthread_mutex_cond_lock (mutex);
  /* XXX Rather abort on errors that are disallowed by POSIX?  */
  return (err != 0) ? err : result;
}

int
__pthread_cond_wait (pthread_cond_t *cond, pthread_mutex_t *mutex)
{
  return __pthread_cond_wait_common (cond, mutex, NULL);
}

int
__pthread_cond_timedwait (pthread_cond_t *cond, pthread_mutex_t *mutex,
    const struct timespec *abstime)
{
  /* Check parameter validity.  This should also tell the compiler that
     it can assume that abstime is not NULL.  */
  if (abstime->tv_nsec < 0 || abstime->tv_nsec >= 1000000000)
    return EINVAL;
  return __pthread_cond_wait_common (cond, mutex, abstime);
}

versioned_symbol (libpthread, __pthread_cond_wait, pthread_cond_wait,
		  GLIBC_2_3_2);
versioned_symbol (libpthread, __pthread_cond_timedwait, pthread_cond_timedwait,
		  GLIBC_2_3_2);
